# Replace common gocr mistakes in dutch language
# Please use info sed to obtain more information
# about sed syntax or read
# http://www.ptug.org/sed/sedfaq.htm
# 2002-09-09: Created by Dominique De Munck.
s/\<lk\>/Ik/g
s/\<ls\>/Is/g
s/\<ln\>/In/g
s/\<lemand\>/Iemand/g
