# Replace common gocr mistakes in english language
# Please use info sed to obtain more information
# about sed syntax or read
# http://www.ptug.org/sed/sedfaq.htm
# 2002-6-18: Modified to use \<..\> word boundaries.
s/\<l\>/I/g
s/\<l'll\>/I'll/g
s/\<ln\>/In/g
s/\<lt\>/It/g
s/\<lt's\>/It's/g
s/\<l've\>/I've/g
s/\<l'd\>/I'd/g
s/\<-l\>/-I/g
s/\<l?\>/I?/g
s/\<iust\>/just/g
s/\<ls\>/Is/g
s/\.iust /.just/g
s/\<lsrael\>/Israel/g
s/\<lgrayne\>/Igrayne/g
s/\<lf\>/If/g
s/\<l'm\>/I'm/g
s/\<lt'd\>/It'd/g

