# Replace common gocr mistakes in frensh language
# Please use info sed to obtain more information
# about sed syntax or read
# http://www.ptug.org/sed/sedfaq.htm
# File contributed by
# Guillaume Rousse <rousse@ccr.jussieu.fr>
s/\<Ies\>/les/g
s/\<Ia\>/la/g
s/-Ia\>/-la/g
s/\<Ià\>/là/g
s/\<Ie\>/le/g
s/-Ie\>/-le/g
s/\<Iui\>/lui/g
s/\<Ieur\>/leur/g
s/\<I '/l'/g
